@@ -193,7 +193,7 @@ def era5_0250_1h(fname, **kwargs):
 
     ds_era5 = xr.open_dataset(fname, chunks={'time': 1})
     ds_era5 = ds_era5[['u10', 'v10', 'latitude025', 'longitude025']]
-    ds_era5 = ds_era5.sel(time=str(kwargs['date']))
+    ds_era5 = ds_era5.sel(time=str(kwargs['date']), method="nearest")
     ds_era5 = ds_era5.drop('time')
 
     ds_era5 = ds_era5.rename(
